open Printf
open Ast
open Ast_typer

let rec print_compil_unit oc {
      package = package;
      imports = imports;
      class_decls = class_decls;
      interf_decls = interf_decls;
    } =
  print_package_option oc package;
  List.iter (print_import oc) imports;
  if imports <> [] then fprintf oc "\n";
  List.iter (print_class_decl oc) class_decls;
  List.iter (print_interf_decl oc) interf_decls;
  flush oc;

and print_package_option oc po =
  match po with
  | Some p -> fprintf oc "package %a;\n\n" print_path p
  | None -> ()

and print_import oc i =
  match i with
  | Isingle l -> fprintf oc "import %a;\n" print_path l
  | Iall p -> fprintf oc "import %a.*;\n" print_path p

and print_slist oc l =
  match l with
  | [] -> ()
  | [ x ] -> fprintf oc "%s" x
  | hd :: tl -> fprintf oc "%s %a" hd print_slist tl

and print_path oc p =
  match p with
  | [] -> ()
  | [ x ] -> fprintf oc "%s" x
  | hd :: tl -> fprintf oc "%s.%a" hd print_path tl

and print_option p oc o =
  match o with
  | Some x -> p oc x
  | None -> ()

and print_extend oc c =
  fprintf oc "extends %a" print_path c;

and print_implements oc l =
  if l <> [] then
    let rec f oc l =
      match l with
      | [] -> ()
      | [ x ] -> print_path oc x
      | hd :: tl -> fprintf oc "%a, %a" print_path hd f tl
    in
      fprintf oc "implements ";
      f oc l;

and print_extends oc l =
  if l <> [] then (
    fprintf oc "extends ";
    List.iter (print_path oc) l;
  )

and string_of_modifier m =
  match m with
  | `abstract -> "abstract"
  | `final -> "final"
  | `native -> "native"
  | `private_ -> "private"
  | `protected -> "protected"
  | `public -> "public"
  | `static -> "static"
  | `synchronized -> "synchronized"
  | `transient -> "transient"
  | `volatile -> "volatile"

and print_modifiers oc m =
  print_slist oc (List.map string_of_modifier m);
  if m <> [] then fprintf oc " ";

and print_arity oc n =
  for i = 1 to n do fprintf oc "[]" done

and print_class_decl oc {
      class_modifiers = class_modifiers;
      class_name = class_name;
      class_extend = class_extend;
      class_implements = class_implements;
      class_fields = class_fields;
      class_methods = class_methods;
      class_initializers = class_initializers;
      class_constructors = class_constructors;
    } =
  fprintf oc "%aclass %s %a %a {\n"
    print_modifiers class_modifiers
    class_name
    print_extend class_extend
    print_implements class_implements;
  List.iter (print_field oc) class_fields;
  fprintf oc "\n";
  List.iter (print_initializer oc) class_initializers;
  fprintf oc "\n";
  List.iter (print_constructor oc) class_constructors;
  fprintf oc "\n";
  List.iter (print_method oc) class_methods;
  fprintf oc "}\n";

and print_interf_decl oc {
      interf_modifiers = interf_modifiers;
      interf_name = interf_name;
      interf_extends = interf_extends;
      interf_constants = interf_constants;
      interf_abstracts = interf_abstracts;
    } =
  fprintf oc "%a interface %s %a {\n"
    print_modifiers interf_modifiers
    interf_name
    print_extends interf_extends;
  List.iter (print_field oc) interf_constants;
  fprintf oc "\n";
  List.iter (fprintf oc "%a\n" print_abstract) interf_abstracts;
  fprintf oc "}\n";

and print_field oc {
      field_modifiers = field_modifiers;
      field_type = field_type;
      field_name = field_name;
      field_initializer = field_initializer;
    } =
  fprintf oc "  %a%a %s"
    print_modifiers field_modifiers
    print_type field_type
    field_name;
  match field_initializer with
  | Some i -> fprintf oc " = %a;\n" print_init_expr i;
  | None -> fprintf oc ";\n";

and print_method oc {
      method_modifiers = method_modifiers;
      method_return = method_return;
      method_name = method_name;
      method_parameters = method_parameters;
      method_throws = method_throws;
      method_body = method_body;
    } =
  fprintf oc "  %a%a %s("
    print_modifiers method_modifiers
    print_type_option method_return
    method_name;
  print_parameters oc method_parameters;
  fprintf oc ")";
  fprintf oc " ";
  print_throws oc method_throws;
  if method_throws <> [] then fprintf oc " ";
  match method_body with
  | None -> fprintf oc ";\n"
  | Some b ->
      fprintf oc "{\n";
      List.iter (print_instruction 2 oc) b;
      fprintf oc "  }\n";

and print_initializer oc b =
  fprintf oc "  static {\n";
  List.iter (print_instruction 2 oc) b;
  fprintf oc "  }\n";

and print_constructor oc {
      constructor_modifiers = constructor_modifiers;
      constructor_name = constructor_name;
      constructor_parameters = constructor_parameters;
      constructor_throws = constructor_throws;
      constructor_call = constructor_call;
      constructor_body = constructor_body;
    } =
  fprintf oc "  %a%s("
    print_modifiers constructor_modifiers
    constructor_name;
  print_parameters oc constructor_parameters;
  fprintf oc ") ";
  print_throws oc constructor_throws;
  if constructor_throws <> [] then fprintf oc " ";
  match constructor_body with
  | None -> fprintf oc ";\n"
  | Some b ->
      fprintf oc "{\n";
      List.iter (print_instruction 2 oc) b;
      fprintf oc "  }\n";

and print_abstract oc {
      abstract_modifiers = abstract_modifiers;
      abstract_return = abstract_return;
      abstract_name = abstract_name;
      abstract_parameters = abstract_parameters;
      abstract_throws = abstract_throws;
    } =
  fprintf oc "  %a%a %s("
    print_modifiers abstract_modifiers
    print_type_option abstract_return
    abstract_name;
  print_parameters oc abstract_parameters;
  fprintf oc ")";
  fprintf oc " ";
  print_throws oc abstract_throws;
  if abstract_throws <> [] then fprintf oc " ";
  fprintf oc ";\n"

and print_parameters oc p =
  let print_parameter (t, n) =
    fprintf oc "%a %s" print_type t n;
  in
  let rec f l =
    match l with
    | [] -> ()
    | [ x ] -> print_parameter x
    | hd :: tl -> print_parameter hd ; fprintf oc ", " ; f tl
  in
    f p

and print_throws oc t =
  if t <> [] then
    let rec f l =
      match l with
      | [] -> ()
      | [ x ] -> print_path oc x
      | hd :: tl -> print_path oc hd ; fprintf oc ", " ; f tl
    in
      fprintf oc "throws ";
      f t;

and string_of_type t =
  match t with
  | `boolean -> "boolean"
  | `byte -> "byte"
  | `char -> "char"
  | `short -> "short"
  | `int -> "int"
  | `long -> "long"
  | `float -> "float"
  | `double -> "double"
  | `string -> "string"
  | `java_object cd ->
      List.fold_left (fun acc x -> acc ^ "." ^ x) (List.hd cd) (List.tl cd)
  | `array t -> string_of_type t ^ "[]"
  | `void -> "void"
  | `nil -> "null"

and print_type oc t =
  fprintf oc "%s" (string_of_type t)

and print_type_option oc o =
  match o with
  | Some t -> print_type oc t
  | None -> fprintf oc "void"

and print_spaces oc indent =
  for i = 1 to indent do fprintf oc "  " done

and print_block indent oc l =
  print_spaces oc indent;
  fprintf oc "{\n";
  List.iter (print_instruction (succ indent) oc) l;
  print_spaces oc indent;
  fprintf oc "}\n";

and print_init_expr oc ie =
  match ie with
  | Ast.Simple_init e -> print_expression oc e
  | Ast.Array_init l ->
      let rec f l =
        match l with
        | [] -> ()
        | [ x ] -> print_init_expr oc x
        | hd :: tl -> print_init_expr oc hd ; fprintf oc ", " ; f tl
      in
        fprintf oc "{";
        f l;
        fprintf oc "}";

and print_arguments oc el =
  let rec f l =
    match l with
    | [] -> ()
    | [ x ] -> print_expression oc x
    | hd :: tl -> print_expression oc hd ; fprintf oc ", " ; f tl
  in
    f el

and print_array_arguments oc el =
  let rec f l =
    match l with
    | [] -> ()
    | [ x ] -> fprintf oc "[%a]" print_expression x
    | hd :: tl -> fprintf oc "[%a]" print_expression hd ; f tl
  in
    f el

and print_affect oc af =
  match af with
  | Ast.Aname sl -> print_path oc sl
  | Ast.Afield (e, s) -> fprintf oc "%a.%s" print_expression e s
  | Ast.Aarray (e0, e1) ->
      fprintf oc "(%a)[%a]" print_expression e0 print_expression e1

and print_expression oc e =
  match e with
  | Ast.Eunary (Ast.Opostinc, e0) ->
      fprintf oc "(%a) ++" print_expression e0
  | Ast.Eunary (Ast.Opostdec, e0) ->
      fprintf oc "(%a) --" print_expression e0
  | Ast.Eunary (op, e0) ->
      fprintf oc "%s (%a)" (string_of_unary_op op) print_expression e0
  | Ast.Ebinary (op, e0, e1) ->
      fprintf oc "(%a) %s (%a)" print_expression e0 (string_of_binary_op op)
        print_expression e1
  | Ast.Ebincomp (op, e0, e1) ->
      fprintf oc "(%a) %s (%a)" print_expression e0 (string_of_binary_comp op)
        print_expression e1
  | Ast.Eternary (Ast.Oifthenelse, e0, e1, e2) ->
      fprintf oc "(%a) ? (%a) : (%a)" print_expression e0 print_expression e1
        print_expression e2
  | Ast.Ebool b -> fprintf oc "%b" b
  | Ast.Ebyte n | Ast.Echar n | Ast.Eshort n | Ast.Eint n -> fprintf oc "%d" n
  | Ast.Efloat f | Ast.Edouble f -> fprintf oc "%f" f
  | Ast.Estring s -> fprintf oc "\"%s\"" s
  | Ast.Enew_object (t, el) ->
      fprintf oc "new %a(%a)" print_path t print_arguments el
  | Ast.Enew_array (t, el, a) ->
      fprintf oc "new %a%a%a" print_type t print_array_arguments el
	print_arity a
  | Ast.Efield_access (e0, s) -> fprintf oc "(%a).%s" print_expression e0 s
  | Ast.Ecall (m, args) ->
      fprintf oc "%s(%a)" m print_arguments args
  | Ast.Eexpr_call (e0, m, args) ->
      fprintf oc "%a.%s(%a)" print_expression e0 m print_arguments args
  | Ast.Earray_access (tab, arg) ->
      fprintf oc "(%a)[%a]" print_expression tab print_expression arg
  | Ast.Etype_cast (t, e0) ->
      fprintf oc "(%a) (%a)" print_type t print_expression e0
  | Ast.Eexpr_cast (t, e0) ->
      fprintf oc "(%a) (%a)" print_expression t print_expression e0
  | Ast.Einstanceof (e0, t) ->
      fprintf oc "(%a) instanceof (%a)" print_expression e0 print_type t
  | Ast.Eaffect (af, e0) ->
      fprintf oc "(%a) = (%a)" print_affect af print_expression e0
  | Ast.Evar s -> fprintf oc "%s" s
  | Ast.Enull -> fprintf oc "null"
  | Ast.Ethis -> fprintf oc "this"
  | Ast.Esuper -> fprintf oc "super"

and print_for_init oc fi =
  match fi with
  | Ast.Fnone -> ()
  | Ast.Fexprs el -> print_arguments oc el
  | Ast.Fdecls dl ->
      let print_decl (t, n, i) =
        fprintf oc "%a %s" print_type t n;
        match i with
        | Some e -> fprintf oc " = %a" print_init_expr e
        | None -> ()
      in
      let rec f l =
        match l with
        | [] -> ()
        | [ d ] -> print_decl d
        | hd :: tl -> print_decl hd ; fprintf oc ", " ; f tl
      in
        f dl

and print_instruction indent oc i =
  print_spaces oc indent;
  match i with
  | Ast.Ilabeled (l, i0) ->
      fprintf oc "%s:\n%a" l (print_instruction (indent + 1)) i0
  | Ast.Iexpr e -> fprintf oc "%a;\n" print_expression e
  | Ast.Idecl (t, n, i) ->
      fprintf oc "%a %s" print_type t n;
      begin match i with
        | Some ie -> fprintf oc " = %a;\n" print_init_expr ie
        | None -> fprintf oc ";\n"
      end
  | Ast.Iblock il ->
      fprintf oc "{\n";
      List.iter (print_instruction (indent + 1) oc) il;
      print_spaces oc indent;
      fprintf oc "}\n";
  | Ast.Iifthenelse (e, i0, i1) ->
      fprintf oc "if(%a){\n" print_expression e;
      print_instruction (indent + 1) oc i0;
      print_spaces oc indent;
      fprintf oc "} else {\n";
      print_instruction (indent + 1) oc i1;
      print_spaces oc indent;
      fprintf oc "}\n";
  | Ast.Iswitch (e, b) ->
      fprintf oc "switch(%a){" print_expression e;
      let print_lb (ls, b) =
        let print_label l =
          match l with
          | Ast.Scase e ->
              fprintf oc "%acase %a:" print_spaces indent print_expression e
          | Ast.Sdefault ->
              fprintf oc "%adefault:" print_spaces indent
        in
          List.iter print_label ls;
          List.iter (print_instruction (indent + 1) oc) b;
      in
        List.iter print_lb b;
        print_spaces oc indent;
        fprintf oc "}\n";
  | Ast.Iwhile (e, i) ->
      fprintf oc "while(%a)\n%a%a\n" print_expression e
        (print_instruction (indent + 1)) i print_spaces indent
  | Ast.Idowhile (i, e) ->
      fprintf oc "do{\n%a%a} while(%a);\n" (print_instruction (indent + 1)) i
        print_spaces indent print_expression e
  | Ast.Ifor (i, t, p, b) ->
      fprintf oc "for(%a ;" print_for_init i;
      begin match t with
        | Some e -> fprintf oc " %a ; " print_expression e
        | None -> fprintf oc "; "
      end;
      print_arguments oc p;
      fprintf oc ")\n";
      print_instruction (indent + 1) oc b;
  | Ast.Ibreak lo ->
      begin match lo with
        | Some l -> fprintf oc "break %s;\n" l
        | None -> fprintf oc "break;\n"
      end
  | Ast.Icontinue lo ->
      begin match lo with
        | Some l -> fprintf oc "continue %s;\n" l
        | None -> fprintf oc "continue;\n"
      end
  | Ast.Ireturn eo ->
      begin match eo with
        | Some e -> fprintf oc "return %a;\n" print_expression e
        | None -> fprintf oc "return;\n"
      end
  | Ast.Ithrow e -> fprintf oc "throw(%a)" print_expression e
  | Ast.Isynchronized (e, b) ->
      fprintf oc "synchronized(%a){\n" print_expression e;
      List.iter (print_instruction (indent + 1) oc) b;
      print_spaces oc indent;
      fprintf oc "}\n";
  | Ast.Itry (b, cl, fo) ->
      fprintf oc "try{\n";
      List.iter (print_instruction (indent + 1) oc) b;
      let print_catch (t, n, b) =
        print_spaces oc indent;
        fprintf oc "}catch(%a %s){\n" print_type t n;
        List.iter (print_instruction (indent + 1) oc) b;
      in
        List.iter print_catch cl;
        begin match fo with
          | Some b ->
              print_spaces oc indent;
              fprintf oc "}finally{\n";
              List.iter (print_instruction (indent + 1) oc) b;
          | None -> ()
        end;
        print_spaces oc indent;
        fprintf oc "}\n";
  | Ast.Inop -> fprintf oc ";\n"
  | Ast.Iprintln e -> fprintf oc "println(%a);\n" print_expression e

let print_typed_exp oc exp =
  let rec pte_rec exp lvl =
    let rec print_unaryop op e =
      print_spaces oc lvl;
      begin match op with
        | Oneg | Onot | Otild
        | Opreinc | Opredec ->
            output_string oc (string_of_unary_op op) ; pte_rec e (lvl + 1)
        | Opostinc | Opostdec ->
            pte_rec e (lvl + 1);
            output_string oc (string_of_unary_op op)
      end;
      fprintf oc "\n";
    and print_cast e =
      print_spaces oc lvl;
      output_string oc " cast ";
      fprintf oc "\n";
      pte_rec e (lvl + 1);
      fprintf oc "\n";
    and print_binaryop op e1 e2 =
      pte_rec e1 (lvl + 1);
      print_spaces oc lvl;
      output_string oc (string_of_binary_op op);
      fprintf oc "\n";
      pte_rec e2 (lvl + 1);
    and print_binarycomp op e1 e2 =
      pte_rec e1 (lvl + 1);
      print_spaces oc lvl;
      output_string oc (string_of_binary_comp op);
      fprintf oc "\n";
      pte_rec e2 (lvl + 1);
    in
      print_spaces oc lvl;
      output_string oc (string_of_java_type (fst exp)) ;
      output_string oc "(";
      
      (* write on a single line or several lines ? *)
      begin match (snd exp) with
        | Ast_typer.Ebool(_) | Ast_typer.Eint(_) | Ast_typer.Efloat(_) | Ast_typer.Estring(_)
        | _ -> fprintf oc "\n"
      end;
      begin match (snd exp) with
        | Ast_typer.Eunary(op, e) -> print_unaryop op e
        | Ast_typer.Ebinary(op, e1, e2) -> print_binaryop op e1 e2
        | Ast_typer.Ebincomp(op, e1, e2) -> print_binarycomp op e1 e2
        | Ast_typer.Ecast (_, e) -> print_cast e
        | Ast_typer.Ebool (b) -> output_string oc (string_of_bool b);
        | Ast_typer.Eint(i) -> output_string oc (string_of_int i);
        | Ast_typer.Efloat (f) -> output_string oc (string_of_float f);
        | Ast_typer.Estring (s) -> output_string oc s;
        | _ -> output_string oc "XXXXX\n"
      end;
      begin match (snd exp) with
        | Ast_typer.Ebool(_) | Ast_typer.Eint(_) | Ast_typer.Efloat(_) | Ast_typer.Estring(_) 
        -> output_string oc " )\n"
        | _ -> print_spaces oc lvl; output_string oc ")\n"
      end;
  in pte_rec exp 0;;

let rec print_typed_compil_unit oc {
      typed_package = typed_package;
      typed_imports = typed_imports;
      typed_class = typed_class;
      typed_interf_decls = typed_interf_decls;
    } =
  fprintf oc "nothing yet...\n";
  flush oc;;

let print_type_error oc e =
  match e with
  | Type_error (UnmanagedExpression(e)) ->
      Printf.printf "Unmanaged expression : \n%a\n" print_expression e;
      failwith "Unmanaged expression"
  | Type_error (UnexpectedType(jt)) ->
      output_string oc ("Unexpected type" ^ (string_of_java_type jt) ^ "\n");
      failwith "type_check"
  | Type_error (Clash(lt1, lt2)) ->
      output_string oc ("Clash : " ^ (string_of_java_type lt1)
          ^ " vs " ^ (string_of_java_type lt2) ^ "\n");
      failwith "type_check"
  | Type_error (InvalidImplicitCast(lt1, lt2)) ->
      output_string oc ("No implicit cast for " ^ (string_of_java_type lt1)
          ^ " into " ^ (string_of_java_type lt2) ^ "\n");
      failwith "type_check"
  | Type_error (InvalidReturnType(t1, t2)) -> 
    output_string oc ("Invalid return type " ^ (string_of_java_type t1)
          ^ " : expected " ^ (string_of_java_type t2) ^ "\n");
      failwith "type_check"
  | Type_error (IncompatibleTypes (lt1, type_list)) ->
      output_string oc ("No compatible type for "
          ^ (string_of_java_type lt1) ^ " in "
          ^ "[ " ^ (string_of_java_types type_list) ^ " ]\n");
      failwith "type_check"
  | Type_error(UndeclaredClass(str))
  -> Printf.printf "Undeclared class %s\n%!" str; failwith "type_check"
  | Type_error (UndeclaredVariable(str))
  -> Printf.printf "Undeclared variable %s\n%!" str ; failwith "type_check"
  | Type_error (UndeclaredMethod(str))
  -> Printf.printf "Undeclared method %s\n%!" str ; failwith "type_check"
  | Type_error (RedeclaredVariableInBlock(str))
  -> Printf.printf "%s is already declared in this block\n%!" str ; failwith "type_check"
  | Type_error (RedeclaredVariableInClass(str))
  -> Printf.printf "%s is already declared in this class\n%!" str ; failwith "type_check"
  | Type_error (NotAnObject(t))
  -> Printf.printf "The following type is not an object : %s\n%!" (string_of_java_type t) ; failwith "type_check"
  | Type_error (RedeclaredVariableInMethod(str))
  -> Printf.printf "%s is the name of another paramer in this method\n%!" str ; failwith "type_check" 
  | Type_error (CallingNonStaticFromStatic(str))
  -> Printf.printf "Calling non static method %s from static method\n%!" str ; failwith "type_check"
  | Type_error (CannotInstantiateAbstractClass(c))
  -> Printf.printf "Cannot instantiate abstract class %s\n%!" c ; failwith "type_check"
  | Type_error (AbstractMethodInClass(c))
  -> Printf.printf "An unimplemented abstract method was found in non abstract class %s\n%!" c ; failwith "type_check"
  | Type_error (AmbiguousCall(algo, s, l))
  -> Printf.printf "Ambiguous call %s using the %s algorithm\nmultiple candidate methods found : \n%!" s algo; List.iter (Printf.printf "  %s\n") l ; failwith "type_check"
  | _ -> failwith "other error";;
